from typing import Any


class HttpStatusError(RuntimeError):
    def __init__(self, status_code: int, status_text: str, body: Any):
        self.status_code = status_code
        self.status_text = status_text
        self.body = body

    def __repr__(self):
        return (
            f"{self.__class__.__name__}"
            f"({self.status_code}, {self.status_text!r}, {self.body!r})"
        )

    __str__ = __repr__


class ClientError(HttpStatusError):
    pass


class ServerError(HttpStatusError):
    pass



__all__ = [
    "ClientError",
    "Dumper",
    "HttpStatusError",
    "JsonRPCBuilder",
    "JsonRPCError",
    "JsonRPCIdMismatchError",
    "Loader",
    "RestBuilder",
    "ServerError",
]

from .client import Dumper, Loader
from .exceptions import ClientError, HttpStatusError, ServerError
from .jsonrpc import JsonRPCBuilder, JsonRPCError, JsonRPCIdMismatchError
from .rest_builder import RestBuilder



from abc import abstractmethod
from dataclasses import dataclass, field
from typing import Any, Protocol, runtime_checkable

from kiss_headers import Headers

from descanso.request import HttpRequest


@dataclass
class HttpResponse:
    status_code: int
    status_text: str
    url: str = ""
    headers: Headers = field(default_factory=Headers)
    body: Any = None


@runtime_checkable
class ResponseTransformer(Protocol):
    @abstractmethod
    def need_response_body(self, response: HttpResponse) -> bool:
        raise NotImplementedError

    @abstractmethod
    def transform_response(
        self,
        request: HttpRequest,
        response: HttpResponse,
    ) -> HttpResponse:
        raise NotImplementedError


class BaseResponseTransformer(ResponseTransformer):
    def need_response_body(self, response: HttpResponse) -> bool:
        return False

    def transform_response(
        self,
        request: HttpRequest,
        response: HttpResponse,
    ) -> HttpResponse:
        return response

    def __or__(self, other: ResponseTransformer) -> "PipeResponseTransformer":
        return PipeResponseTransformer(self, other)

    def __ror__(self, other: ResponseTransformer) -> "PipeResponseTransformer":
        return PipeResponseTransformer(other, self)


class PipeResponseTransformer(BaseResponseTransformer):
    def __init__(self, *others: ResponseTransformer) -> None:
        self.others = others

    def need_response_body(self, response: HttpResponse) -> bool:
        return any(other.need_response_body(response) for other in self.others)

    def transform_response(
        self,
        request: HttpRequest,
        response: HttpResponse,
    ) -> HttpResponse:
        for other in self.others:
            response = other.transform_response(request, response)
        return response



from collections.abc import Awaitable, Callable
from typing import (
    Any,
    Concatenate,
    Generic,
    ParamSpec,
    TypeVar,
    overload,
)

from .bound_method import BoundAsyncMethod, BoundSyncMethod
from .client import AsyncClient, SyncClient
from .method_spec import MethodSpec

_MethodResultT = TypeVar("_MethodResultT")
_MethodParamSpec = ParamSpec("_MethodParamSpec")


class MethodBinder(Generic[_MethodParamSpec, _MethodResultT]):
    def __init__(
        self,
        spec: MethodSpec[Concatenate[Any, _MethodParamSpec], _MethodResultT],
    ) -> None:
        self._spec = spec

    @property
    def spec(
        self,
    ) -> MethodSpec[Concatenate[Any, _MethodParamSpec], _MethodResultT]:
        return self._spec

    @overload
    def __get__(
        self,
        instance: SyncClient,
        owner: Any = None,
    ) -> Callable[_MethodParamSpec, _MethodResultT]: ...

    @overload
    def __get__(
        self,
        instance: AsyncClient,
        owner: Any = None,
    ) -> Callable[_MethodParamSpec, Awaitable[_MethodResultT]]: ...

    @overload
    def __get__(
        self,
        instance: None,
        owner: Any = None,
    ) -> MethodSpec[_MethodParamSpec, _MethodResultT]: ...

    def __get__(
        self,
        instance: Any,
        owner: Any = None,
    ) -> Any:
        if instance is None:
            return self
        elif isinstance(instance, SyncClient):
            return BoundSyncMethod(self._spec, instance)
        elif isinstance(instance, AsyncClient):
            return BoundAsyncMethod(self._spec, instance)
        else:
            raise TypeError



import json
from collections.abc import Sequence
from typing import Any

from .client import Loader
from .exceptions import ClientError, ServerError
from .request import HttpRequest
from .response import BaseResponseTransformer, HttpResponse


class BodyModelLoad(BaseResponseTransformer):
    def __init__(
        self,
        type_hint: Any,
        loader: Loader | None,
    ) -> None:
        self.type_hint = type_hint
        self.loader = loader

    def transform_response(
        self,
        request: HttpRequest,
        response: HttpResponse,
    ) -> HttpResponse:
        response.body = self.loader.load(response.body, self.type_hint)
        return response

    def __repr__(self):
        return (
            f"{self.__class__.__name__}({self.type_hint!r}, {self.loader!r})"
        )


class JsonLoad(BaseResponseTransformer):
    def __init__(self, codes: Sequence[int] = (200, 201, 202)):
        self.codes = codes

    def need_response_body(self, response: HttpResponse) -> bool:
        return response.status_code in self.codes

    def transform_response(
        self,
        request: HttpRequest,
        response: HttpResponse,
    ) -> HttpResponse:
        if response.status_code not in self.codes:
            return response

        response.body = json.loads(response.body)
        return response

    def __repr__(self):
        return f"{self.__class__.__name__}()"


class ErrorRaiser(BaseResponseTransformer):
    def __init__(
        self,
        *,
        codes: Sequence[int] | None = None,
        except_codes: Sequence[int] | None = None,
        need_body: bool = False,
    ) -> None:
        self._need_body = need_body
        self.codes = codes
        self.except_codes = except_codes

    def need_response_body(self, response: HttpResponse) -> bool:
        return self._need_body

    def transform_response(
        self,
        request: HttpRequest,
        response: HttpResponse,
    ) -> HttpResponse:
        if not self._is_status_code_allowed(response.status_code):
            if response.status_code >= 500:  # noqa: PLR2004
                raise ServerError(
                    status_code=response.status_code,
                    status_text=response.status_text,
                    body=response.body,
                )

            raise ClientError(
                status_code=response.status_code,
                status_text=response.status_text,
                body=response.body,
            )
        return response

    def _is_status_code_allowed(self, code: int) -> bool:
        if self.except_codes and code not in self.except_codes:
            return False

        if self.codes and code in self.codes:
            return False

        if not self.codes and not self.except_codes and code >= 400:  # noqa: PLR2004
            return False

        return True

    def __repr__(self):
        return f"{self.__class__.__name__}({self.codes!r}, {self._need_body})"


class KeepResponse(BaseResponseTransformer):
    def __init__(self, *, need_body: bool):
        self._need_body = need_body

    def need_body(self, response: HttpResponse) -> bool:
        return self._need_body

    def transform_response(
        self,
        request: HttpRequest,
        response: HttpResponse,
    ) -> HttpResponse:
        return HttpResponse(
            url=response.url,
            status_code=response.status_code,
            status_text=response.status_text,
            headers=response.headers,
            body=response,
        )

    def __repr__(self):
        return f"{self.__class__.__name__}(need_body={self._need_body})"



# This file auto generated!
# DO NOT EDIT MANUALLY!

version: str = "0.7.0"
__version__: str = "0.7.0"

commit_hash: str = "g6a1b2b50a6835fe170fb5e9036e22537650f3a5d"
version_timestamp: str = "2025-11-29 23:56:07.377421+00:00"

tag: str = "0.7.0"
branch: str = "HEAD"
commit_date: str = "2025-11-29"
commit_count_since_tag: int = 0



from collections.abc import Awaitable, Callable
from typing import (
    Any,
    Concatenate,
    ParamSpec,
    TypedDict,
    TypeVar,
    overload,
)

try:
    from typing import Unpack
except ImportError:
    T = TypeVar("T")
    Unpack = Any | T

from descanso import Dumper, Loader
from descanso.builder_base import (
    DEFAULT_BODY_PARAM,
    Decorator,
    Transformer,
    UrlSrc,
    url_transformer,
)
from descanso.method_descriptor import MethodBinder
from descanso.method_spec import MethodSpec
from descanso.request import FieldDestination, FieldOut, RequestTransformer
from descanso.request_transformers import (
    Body,
    BodyModelDump,
    FormQuery,
    JsonDump,
    Method,
    Query,
    QueryModelDump,
)
from descanso.response import HttpResponse, ResponseTransformer
from descanso.response_transformers import (
    BodyModelLoad,
    ErrorRaiser,
    JsonLoad,
    KeepResponse,
)
from descanso.signature import make_method_spec

_MethodResultT = TypeVar("_MethodResultT")
_MethodParamSpec = ParamSpec("_MethodParamSpec")


class BuilderParams(TypedDict):
    body_name: str

    query_param_dumper: Dumper | None
    request_body_dumper: Dumper | None
    request_body_post_dump: RequestTransformer | None
    query_param_post_dump: RequestTransformer | None

    response_body_loader: Loader | None
    response_body_pre_load: ResponseTransformer | None
    error_raiser: ResponseTransformer | None


class RestBuilder(Decorator):
    def __init__(
        self,
        *transformers: Transformer,
        **params: Unpack[BuilderParams],
    ) -> None:
        self.transformers = transformers
        self.params = params

    def with_params(
        self,
        *transformers: Transformer,
        **params: Unpack[BuilderParams],
    ) -> "RestBuilder":
        return RestBuilder(
            *transformers,
            *self.transformers,
            **(self.params | params),
        )

    def get(
        self,
        url: UrlSrc,
        *transformers: Transformer,
        **params: Unpack[BuilderParams],
    ) -> "RestBuilder":
        return self.with_params(
            url_transformer(url),
            Method("GET"),
            *transformers,
            **params,
        )

    def post(
        self,
        url: UrlSrc,
        *transformers: Transformer,
        **params: Unpack[BuilderParams],
    ) -> "RestBuilder":
        return self.with_params(
            url_transformer(url),
            Method("POST"),
            *transformers,
            **params,
        )

    def put(
        self,
        url: UrlSrc,
        *transformers: Transformer,
        **params: Unpack[BuilderParams],
    ) -> "RestBuilder":
        return self.with_params(
            url_transformer(url),
            Method("PUT"),
            *transformers,
            **params,
        )

    def patch(
        self,
        url: UrlSrc,
        *transformers: Transformer,
        **params: Unpack[BuilderParams],
    ) -> "RestBuilder":
        return self.with_params(
            url_transformer(url),
            Method("PATCH"),
            *transformers,
            **params,
        )

    def delete(
        self,
        url: UrlSrc,
        *transformers: Transformer,
        **params: Unpack[BuilderParams],
    ) -> "RestBuilder":
        return self.with_params(
            url_transformer(url),
            Method("DELETE"),
            *transformers,
            **params,
        )

    def _add_request_transformer(
        self,
        spec: MethodSpec,
        transformer: RequestTransformer,
    ):
        spec.request_transformers.append(transformer)
        spec.fields_out.extend(transformer.transform_fields(spec.fields_in))

    def _get_body_field(self, spec: MethodSpec) -> FieldOut | None:
        for field in spec.fields_out:
            if field.dest is FieldDestination.BODY:
                return field
        return None

    def _add_default_request_body_transformers(self, spec: MethodSpec):
        default_body_name = self.params.get("body_name", DEFAULT_BODY_PARAM)

        body_out = self._get_body_field(spec)
        for field in spec.fields_in:
            if field.consumed_by:
                continue
            if not body_out and field.name == default_body_name:
                self._add_request_transformer(spec, Body(field.name))

        if self._get_body_field(spec):
            dumper = self.params.get("request_body_dumper")
            if dumper:
                self._add_request_transformer(spec, BodyModelDump(dumper))

            post_dump = self.params.get("request_body_post_dump", ...)
            if post_dump is ...:
                self._add_request_transformer(spec, JsonDump())
            elif post_dump:
                self._add_request_transformer(spec, post_dump)

    def _add_default_query_transformers(self, spec: MethodSpec):
        for field in spec.fields_in:
            if field.consumed_by:
                continue
            self._add_request_transformer(spec, Query(field.name))

        if dumper := self.params.get("query_param_dumper"):
            self._add_request_transformer(spec, QueryModelDump(dumper))
        query_post_dump = self.params.get("query_param_post_dump", ...)
        if query_post_dump is ...:
            self._add_request_transformer(spec, FormQuery())
        elif query_post_dump:
            self._add_request_transformer(spec, query_post_dump)
        return []

    def _add_default_response_transformers(self, spec: MethodSpec) -> None:
        error_raiser = self.params.get("error_raiser", ...)
        if error_raiser is ...:
            spec.response_transformers.append(ErrorRaiser())
        elif error_raiser:
            spec.response_transformers.append(error_raiser)

        pre_loader = self.params.get("response_body_pre_load", ...)
        if pre_loader is ...:
            spec.response_transformers.append(JsonLoad())
        elif pre_loader:
            spec.response_transformers.append(pre_loader)

        loader = self.params.get("response_body_loader")
        if spec.result_type is HttpResponse:
            spec.response_transformers.append(KeepResponse(need_body=False))
        elif (
            loader
            and spec.result_type is not Any
            and spec.result_type is not object
        ):
            spec.response_transformers.append(
                BodyModelLoad(spec.result_type, loader=loader),
            )

    @overload
    def __call__(
        self,
        func: Callable[
            Concatenate[Any, _MethodParamSpec],
            Awaitable[_MethodResultT],
        ],
    ) -> MethodBinder[_MethodParamSpec, _MethodResultT]: ...

    @overload
    def __call__(
        self,
        func: Callable[Concatenate[Any, _MethodParamSpec], _MethodResultT],
    ) -> MethodBinder[_MethodParamSpec, _MethodResultT]: ...

    def __call__(
        self,
        func: Callable[Concatenate[Any, _MethodParamSpec], _MethodResultT],
    ) -> MethodBinder[_MethodParamSpec, _MethodResultT]:
        spec = make_method_spec(
            func,
            transformers=self.transformers,
            is_in_class=True,
        )
        self._add_default_request_body_transformers(spec)
        self._add_default_query_transformers(spec)
        self._add_default_response_transformers(spec)
        return MethodBinder(spec)



import logging

from pydantic import BaseModel, ConfigDict, TypeAdapter
from requests import Session

from descanso import delete, get, post
from descanso.client import Dumper, Loader
from descanso.http.requests import RequestsClient
from descanso.rest_builder import RestBuilder


class PydanticAdapter(Loader, Dumper):
    def load(self, data, type_):
        return TypeAdapter(type_).validate_python(data)

    def dump(self, data, type_):
        return TypeAdapter(type_).dump_python(data)


def to_camel(string: str) -> str:
    words = string.split("_")
    return words[0] + "".join(word.capitalize() for word in words[1:])


class Todo(BaseModel):
    id: int
    user_id: int
    title: str
    completed: bool

    model_config = ConfigDict(
        alias_generator=to_camel,
        populate_by_name=True,
    )

adapter = PydanticAdapter()
rest = RestBuilder(
    request_body_dumper=adapter,
    response_body_loader=adapter,
    query_param_dumper=adapter,
)


class RealClient(RequestsClient):
    def __init__(self):
        super().__init__(
            base_url="https://jsonplaceholder.typicode.com/",
            session=Session(),
        )

    @rest.get("todos/{id}")
    def get_todo(self, id: str) -> Todo:
        """GET method with path param"""

    @rest.get("todos")
    def list_todos(self, user_id: int | None) -> list[Todo]:
        """GET method with query params"""

    @rest.delete("todos/{id}")
    def delete_todo(self, id: int):
        """DELETE method"""

    @rest.post("todos")
    def create_todo(self, body: Todo) -> Todo:
        """POST method"""

    @rest.get("https://httpbin.org/get")
    def get_httpbin(self):
        """Url different from base_url"""


logging.basicConfig(level=logging.INFO)
client = RealClient()
print(client.list_todos(user_id=1))
print(client.get_todo(id="1"))
print(client.delete_todo(1))
print(client.create_todo(Todo(
    id=123456789, user_id=111222333, title="By Tishka17", completed=False,
)))
print(client.get_httpbin())



from aiohttp import web


async def query_xxy(request: web.Request) -> web.Response:
    assert request.query_string == "x=1&x=2&y=3"
    return web.Response(text="ok")


async def conflict(request: web.Request) -> web.Response:
    return web.Response(
        status=409,
        reason="test_conflict",
        body="test_conflict_body",
    )


async def headers(request: web.Request) -> web.Response:
    assert request.headers["X-Test"] == "1"
    return web.Response(
        headers=[
            ("X-Test", "2"),
            ("X-Test", "3"),
            ("Another", "value"),
        ],
    )


async def delete(request: web.Request) -> web.Response:
    return web.Response(status=204)


async def files(request: web.Request) -> web.Response:
    data = await request.post()
    f1: web.FileField = data["x1"]
    assert f1.filename == "x1"
    assert f1.file.read() == b"1"

    f2: web.FileField = data["x2"]
    assert f2.filename == "x2"
    assert f2.file.read() == b"2"

    f3: web.FileField = data["x3"]
    assert f3.filename == "z3"
    assert f3.content_type == "test_content_type"
    assert f3.file.read() == b"3"
    return web.Response()


async def form(request: web.Request) -> web.Response:
    data = await request.post()
    assert data["field1"] == "1"
    assert data["field2"] == "2"
    return web.Response()


async def json(request: web.Request) -> web.Response:
    data = await request.json()
    assert data == {"x": 1}
    return web.json_response({"y": 2})


async def jsonrpc(request: web.Request) -> web.Response:
    data = await request.json()
    request_id = data["id"]
    assert data["jsonrpc"] == "2.0"
    if data["method"] == "good":
        assert data["params"] == [42]
        return web.json_response(
            {
                "id": request_id,
                "jsonrpc": "2.0",
                "result": 42,
            },
        )
    elif data["method"] == "bad":
        return web.json_response(
            {
                "id": request_id,
                "jsonrpc": "2.0",
                "error": {
                    "code": -32000,
                    "message": "My error",
                },
            },
        )
    else:
        return web.json_response(
            {
                "id": request_id,
                "jsonrpc": "2.0",
                "error": {
                    "code": -32601,
                    "message": "Method not found",
                },
            },
        )


def new_app() -> web.Application:
    app = web.Application()
    app.add_routes(
        [
            web.get("/query_xxy", query_xxy),
            web.get("/conflict", conflict),
            web.get("/headers", headers),
            web.get("/json", json),
            web.delete("/delete", delete),
            web.post("/files", files),
            web.post("/form", form),
            web.post("/jsonrpc", jsonrpc),
        ],
    )
    return app


async def new_site(port: int) -> web.TCPSite:
    app = new_app()
    app_runner = web.AppRunner(app)
    await app_runner.setup()
    return web.TCPSite(app_runner, "localhost", port)



from descanso.request import FieldIn, RequestTransformer


def consumed_fields(
    fields: list[FieldIn],
    transformer: RequestTransformer,
) -> list[str]:
    return [f.name for f in fields if transformer in f.consumed_by]



import logging
import os
from dataclasses import dataclass
from enum import Enum
from typing import Generic, TypeVar

from adaptix import Chain, Retort, dumper
from requests import Session

from descanso.http.requests import RequestsClient
from descanso.request_transformers import Query, DelimiterQuery
from descanso.rest_builder import RestBuilder

T = TypeVar("T")


@dataclass
class Response(Generic[T]):
    response: T


@dataclass
class User:
    id: int
    first_name: str
    last_name: str


class GenderQuery(Enum):
    MALE = 2
    FEMALE = 1
    ANY = 9


@dataclass
class UsersSearchResult:
    count: int
    items: list[User]

retort = Retort()
query_dumper = Retort(recipe=[
    dumper(bool, int),
    dumper(int, str),
    dumper(list[int], lambda d: ",".join(d), Chain.LAST),
    dumper(list[str], lambda d: ",".join(d), Chain.LAST),
])
rest = RestBuilder(
    Query("v", "5.131"),
    request_body_dumper=retort,
    response_body_loader=retort,
    query_param_dumper=retort,
    query_param_post_dump=DelimiterQuery(),
)

class VkClient(RequestsClient):
    def __init__(self, token: str):
        self.token = token
        super(VkClient, self).__init__(
            base_url="https://api.vk.com/method/",
            session=Session(),
            transformers=[
                Query("access_token", "{self.token}")
            ]
        )

    def set_token(self, token: str) -> None:
        self.token = token

    @rest.get("users.get")
    def get_users(self, user_ids: list[str]) -> Response[list[User]]:
        """Get users by their ids"""

    @rest.get("users.search")
    def search_users(
            self, q: str, sort: bool = False,
            offset: int = 0, count: int = 20,
            gender: GenderQuery = GenderQuery.ANY,
    ) -> Response[UsersSearchResult]:
        """Search users with pagination"""


TOKEN = os.getenv("VK_TOKEN")


def main():
    logging.basicConfig(level=logging.INFO)
    client = VkClient(TOKEN)
    print(client.get_users(["1", "2"]))
    print(client.search_users(q="tishka17", gender=GenderQuery.MALE))


if __name__ == "__main__":
    main()



import urllib.parse
from collections.abc import AsyncIterator, Sequence
from contextlib import asynccontextmanager

from aiohttp import ClientResponse, ClientSession, FormData
from kiss_headers import parse_it

from descanso.client import (
    AsyncClient,
    AsyncResponseWrapper,
)
from descanso.request import HttpRequest, RequestTransformer
from descanso.response import ResponseTransformer
from descanso.utils import ensure_trailing_slash


class AiohttpResponseWrapper(AsyncResponseWrapper):
    def __init__(self, response: ClientResponse) -> None:
        self.status_code = response.status
        self.status_text = response.reason
        self.body = None
        self.headers = parse_it(response.headers)
        self._raw_response = response

    async def aload_body(self) -> None:
        self.body = await self._raw_response.read()


class AiohttpClient(AsyncClient):
    def __init__(
        self,
        base_url: str,
        session: ClientSession,
        transformers: Sequence[RequestTransformer | ResponseTransformer] = (),
    ) -> None:
        super().__init__(
            transformers=transformers,
        )
        self._base_url = ensure_trailing_slash(base_url)
        self._session = session

    @asynccontextmanager
    async def asend_request(
        self,
        request: HttpRequest,
    ) -> AsyncIterator[AsyncResponseWrapper]:
        data = request.body
        if request.files:
            data = FormData(data or {})
            for name, file in request.files:
                data.add_field(
                    name,
                    filename=file.filename,
                    content_type=file.content_type,
                    value=file.contents,
                )

        async with self._session.request(
            method=request.method,
            url=urllib.parse.urljoin(self._base_url, request.url),
            headers=request.headers,
            data=data,
            params=[(k, v) for k, v in request.query_params if v is not None],
        ) as resp:
            yield AiohttpResponseWrapper(resp)



from collections.abc import Awaitable, Callable
from typing import (
    Any,
    Concatenate,
    ParamSpec,
    Protocol,
    TypeVar,
    overload,
)

try:
    from typing import Unpack
except ImportError:
    T = TypeVar("T")
    Unpack = Any | T

from descanso.method_descriptor import MethodBinder
from descanso.request import RequestTransformer
from descanso.request_transformers import (
    Url,
)
from descanso.response import ResponseTransformer

DEFAULT_BODY_PARAM = "body"

_MethodResultT = TypeVar("_MethodResultT")
_MethodParamSpec = ParamSpec("_MethodParamSpec")


Transformer = RequestTransformer | ResponseTransformer
UrlSrc = str | Callable | Url


def url_transformer(url: UrlSrc) -> Url:
    if isinstance(url, Url):
        return url
    return Url(url)


class Decorator(Protocol):
    @overload
    def __call__(
        self,
        func: Callable[
            Concatenate[Any, _MethodParamSpec],
            Awaitable[_MethodResultT],
        ],
    ) -> MethodBinder[_MethodParamSpec, _MethodResultT]: ...

    @overload
    def __call__(
        self,
        func: Callable[Concatenate[Any, _MethodParamSpec], _MethodResultT],
    ) -> MethodBinder[_MethodParamSpec, _MethodResultT]: ...

    def __call__(self, *args, **kwargs):
        raise NotImplementedError



from contextlib import AbstractContextManager
from unittest.mock import Mock

from dirty_equals import DirtyEquals, HasAttributes, IsInstance

from descanso.client import SyncClient, SyncResponseWrapper
from descanso.request import HttpRequest


class _Dirty:
    def __getitem__(self, item: type):
        def check(**attrs) -> DirtyEquals:
            if not attrs:
                return IsInstance(item)
            return IsInstance(item) & HasAttributes(**attrs)

        return check


dirty = _Dirty()


class MockClient(SyncClient):
    def __init__(self, mock: Mock) -> None:
        self._mock = mock

    def send_request(
        self,
        request: HttpRequest,
    ) -> AbstractContextManager[SyncResponseWrapper]:
        return self._mock(request)



from typing import Any

from dirty_equals import Contains

from descanso import Loader, RestBuilder
from descanso.client import Dumper
from descanso.request_transformers import (
    Body,
    BodyModelDump,
    FormQuery,
    JsonDump,
    Method,
    Query,
    QueryModelDump,
    Skip,
    Url,
)
from descanso.response_transformers import BodyModelLoad, ErrorRaiser, JsonLoad
from .utils import dirty


class Model:
    pass


def test_get_with_query():
    rest = RestBuilder()

    class Api:
        @rest.get("/foo")
        def do_get(self, x: int) -> Model:
            """Hello"""

    assert Api.do_get.spec.name == "do_get"
    assert Api.do_get.spec.doc == "Hello"
    assert Api.do_get.spec.request_transformers == [
        dirty[Url](original_template="/foo"),
        dirty[Method](method="GET"),
        dirty[Query](name_out="x", original_template=None),
        dirty[FormQuery](),
    ]
    assert Api.do_get.spec.response_transformers == [
        dirty[ErrorRaiser](),
        dirty[JsonLoad](),
    ]


def test_post_with_url_body():
    rest = RestBuilder()

    class Api:
        @rest.post("/foo{x}")
        def do_post(self, x: int, body: str) -> Model:
            """Hello"""

    assert Api.do_post.spec.request_transformers == [
        dirty[Url](original_template="/foo{x}"),
        dirty[Method](method="POST"),
        dirty[Body](arg="body"),
        dirty[JsonDump](),
        dirty[FormQuery](),
    ]
    assert Api.do_post.spec.response_transformers == [
        dirty[ErrorRaiser](),
        dirty[JsonLoad](),
    ]


def test_methods():
    rest = RestBuilder()

    class Api:
        @rest.get("/")
        def do_get(self) -> None: ...

        @rest.post("/")
        def do_post(self) -> None: ...

        @rest.delete("/")
        def do_delete(self) -> None: ...

        @rest.put("/")
        def do_put(self) -> None: ...

        @rest.patch("/")
        def do_patch(self) -> None: ...

    assert Api.do_get.spec.request_transformers == Contains(
        dirty[Method](method="GET"),
    )
    assert Api.do_post.spec.request_transformers == Contains(
        dirty[Method](method="POST"),
    )
    assert Api.do_delete.spec.request_transformers == Contains(
        dirty[Method](method="DELETE"),
    )
    assert Api.do_put.spec.request_transformers == Contains(
        dirty[Method](method="PUT"),
    )
    assert Api.do_patch.spec.request_transformers == Contains(
        dirty[Method](method="PATCH"),
    )


class StubConverter(Dumper, Loader):
    def dump(self, data: Any, class_: Any) -> Any: ...

    def load(self, data: Any, class_: Any) -> Any: ...


def test_params():
    req_additional = Skip("0")
    resp_additional = ErrorRaiser(codes=[200])
    query_param_dumper = StubConverter()
    query_param_post_dump = Skip("1")
    request_body_dumper = StubConverter()
    response_body_loader = StubConverter()
    request_body_post_dump = Skip("2")
    error_raiser = ErrorRaiser(codes=[400])
    response_body_pre_load = ErrorRaiser(codes=[500])
    rest = RestBuilder(
        req_additional,
        resp_additional,
        query_param_dumper=query_param_dumper,
        query_param_post_dump=query_param_post_dump,
        request_body_dumper=request_body_dumper,
        response_body_loader=response_body_loader,
        request_body_post_dump=request_body_post_dump,
        body_name="x",
        error_raiser=error_raiser,
        response_body_pre_load=response_body_pre_load,
    )

    class Api:
        @rest.get("/")
        def do_get(self, x: int, body: str) -> Model: ...

    assert Api.do_get.spec.request_transformers == [
        dirty[Url](original_template="/"),
        dirty[Method](method="GET"),
        req_additional,
        dirty[Body](arg="x"),
        dirty[BodyModelDump](dumper=request_body_dumper),
        request_body_post_dump,
        dirty[Query](name_out="body", original_template=None),
        dirty[QueryModelDump](dumper=query_param_dumper),
        query_param_post_dump,
    ]
    assert Api.do_get.spec.response_transformers == [
        resp_additional,
        error_raiser,
        response_body_pre_load,
        dirty[BodyModelLoad](type_hint=Model, loader=response_body_loader),
    ]


def test_override_params():
    error_raiser = ErrorRaiser(codes=[400])
    response_body_pre_load = ErrorRaiser(codes=[500])
    response_body_pre_load2 = ErrorRaiser(codes=[502])
    rest = RestBuilder(
        error_raiser=error_raiser,
        response_body_pre_load=response_body_pre_load,
    )

    class Api:
        @rest.get("/", response_body_pre_load=response_body_pre_load2)
        def do_get(self, x: int, body: str) -> Model: ...

    assert Api.do_get.spec.response_transformers == [
        error_raiser,
        response_body_pre_load2,
    ]



from abc import abstractmethod
from collections.abc import Sequence
from dataclasses import dataclass, field
from enum import Enum
from typing import (
    IO,
    Any,
    Protocol,
    TypeAlias,
    TypeVar,
    runtime_checkable,
)

from kiss_headers import Headers

T = TypeVar("T")
KeyValue: TypeAlias = tuple[str, T]
KeyValueList: TypeAlias = list[KeyValue[T]]


@dataclass
class FileData:
    contents: str | IO | None | bytes
    content_type: str | None = None
    filename: str | None = None


@dataclass
class HttpRequest:
    body: Any = None
    files: KeyValueList[FileData] = field(default_factory=list)
    query_params: KeyValueList[Any] = field(default_factory=list)
    headers: Headers = field(default_factory=Headers)
    extras: KeyValueList[Any] = field(default_factory=list)
    url: str = ""
    method: str = "GET"


class FieldDestination(Enum):
    URL = "url"
    HEADER = "headers"
    BODY = "body"
    FILE = "files"
    QUERY = "query_params"
    EXTRA = "extras"
    UNDEFINED = "undefined"


@dataclass
class FieldIn:
    name: str
    type_hint: Any
    consumed_by: list["RequestTransformer"] = field(default_factory=list)


@dataclass
class FieldOut:
    name: str | None
    dest: FieldDestination
    type_hint: Any


@runtime_checkable
class RequestTransformer(Protocol):
    @abstractmethod
    def transform_fields(
        self,
        fields_in: Sequence[FieldIn],
    ) -> Sequence[FieldOut]:
        raise NotImplementedError

    @abstractmethod
    def transform_request(
        self,
        request: HttpRequest,
        fields_in: Sequence[FieldIn],
        fields_out: Sequence[FieldOut],
        data: dict[str, Any],
    ) -> HttpRequest:
        raise NotImplementedError


class BaseRequestTransformer(RequestTransformer):
    def transform_fields(
        self,
        fields_in: Sequence[FieldIn],
    ) -> Sequence[FieldOut]:
        return []

    def transform_request(
        self,
        request: HttpRequest,
        fields_in: Sequence[FieldIn],
        fields_out: Sequence[FieldOut],
        data: dict[str, Any],
    ) -> HttpRequest:
        return request

    def __or__(self, other: RequestTransformer) -> "PipeRequestTransformer":
        return PipeRequestTransformer(self, other)

    def __ror__(self, other: RequestTransformer) -> "PipeRequestTransformer":
        return PipeRequestTransformer(other, self)


class PipeRequestTransformer(BaseRequestTransformer):
    def __init__(self, *others: RequestTransformer) -> None:
        self.others = others

    def transform_fields(
        self,
        fields_in: Sequence[FieldIn],
    ) -> Sequence[FieldOut]:
        res = []
        for other in self.others:
            res.extend(other.transform_fields(fields_in))
        return res

    def transform_request(
        self,
        request: HttpRequest,
        fields_in: Sequence[FieldIn],
        fields_out: Sequence[FieldOut],
        data: dict[str, Any],
    ) -> HttpRequest:
        for other in self.others:
            request = other.transform_request(
                request,
                fields_in,
                fields_out,
                data,
            )
        return request



import asyncio
import logging
from dataclasses import dataclass
from io import BytesIO
from typing import Any

import aiohttp
from adaptix import NameStyle, Retort, name_mapping
from aiohttp import ClientSession

from descanso.http.aiohttp import AiohttpClient
from descanso.request_transformers import File
from descanso.rest_builder import RestBuilder


@dataclass
class Todo:
    id: int
    user_id: int
    title: str
    completed: bool


retort = Retort(recipe=[
    name_mapping(name_style=NameStyle.CAMEL),
])
rest = RestBuilder(
    request_body_dumper=retort,
    response_body_loader=retort,
    query_param_dumper=Retort(),
)

class RealAsyncClient(AiohttpClient):
    def __init__(self, session: ClientSession):
        super().__init__(
            base_url="https://jsonplaceholder.typicode.com/",
            session=session,
        )

    @rest.get("todos/{id}")
    async def get_todo(self, id: str) -> Todo:
        """GET method with path param"""

    @rest.get("todos")
    async def list_todos(self, user_id: int | None) -> list[Todo]:
        """GET method with query params"""

    @rest.delete("todos/{id}")
    async def delete_todo(self, id: int):
        """DELETE method"""

    @rest.post("todos")
    async def create_todo(self, body: Todo) -> Todo:
        """POST method"""

    @rest.get("https://httpbin.org/get")
    async def get_httpbin(self) -> Any:
        """Url different from base_url"""

    @rest.post(
        "https://httpbin.org/post",
        File("file"),
    )
    async def upload_image(self, file: BytesIO):
        """Sending binary data"""


async def main():
    async with aiohttp.ClientSession() as session:
        logging.basicConfig(level=logging.DEBUG)
        client = RealAsyncClient(session)
        print(RealAsyncClient.list_todos)
        print()
        print(await client.list_todos(user_id=1))
        print(await client.get_todo(id="1"))
        print(await client.delete_todo(1))
        print(await client.create_todo(
            Todo(123456789, 111222333, "By Tishka17", False)))

        with open("example.py", "rb") as f:
            print(await client.upload_image(f))


loop = asyncio.get_event_loop()
loop.run_until_complete(main())



from collections.abc import Awaitable, Callable, Sequence
from typing import (
    Any,
    Concatenate,
    ParamSpec,
    Protocol,
    TypedDict,
    TypeVar,
    overload,
)
from uuid import uuid4

try:
    from typing import Unpack
except ImportError:
    T = TypeVar("T")
    Unpack = Any | T

from descanso import Dumper, Loader
from descanso.builder_base import (
    Transformer,
    UrlSrc,
    url_transformer,
)
from descanso.method_descriptor import MethodBinder
from descanso.method_spec import MethodSpec
from descanso.request import (
    FieldDestination,
    FieldIn,
    FieldOut,
    HttpRequest,
    RequestTransformer,
)
from descanso.request_transformers import (
    Body,
    BodyModelDump,
    JsonDump,
    Method,
)
from descanso.response import HttpResponse, ResponseTransformer
from descanso.response_transformers import (
    BodyModelLoad,
    ErrorRaiser,
    JsonLoad,
    KeepResponse,
)
from descanso.signature import make_method_spec

_MethodResultT = TypeVar("_MethodResultT")
_MethodParamSpec = ParamSpec("_MethodParamSpec")

EXTRA_JSON_RPC_REQUEST_ID = "JsonRPC.request_id"
EXTRA_JSON_RPC_METHOD = "JsonRPC.method"


def get_extra(request: HttpRequest, expected_key: str) -> Any:
    for key, value in request.extras:
        if key == expected_key:
            return value
    return None


class BaseJsonRPCError(Exception):
    pass


class MultipleBodyError(BaseJsonRPCError):
    def __init__(self, existing: str, new: str):
        self.existing = existing
        self.new = new

    def __str__(self):
        return (
            f"Cannot have multiple body fields, "
            f"found {self.existing}, {self.new}"
        )


class JsonRPCIdMismatchError(BaseJsonRPCError):
    pass


class JsonRPCError(BaseJsonRPCError):
    def __init__(self, code: int, message: str, data: Any) -> None:
        self.code = code
        self.message = message
        self.data = data

    def __repr__(self):
        return f"JsonRPCError({self.code}, {self.message!r}, {self.data!r})"


class IdGenerator(Protocol):
    def __call__(self) -> str: ...


class JsonRPCIdGenerator(RequestTransformer):
    def __init__(self, id_generator: IdGenerator | None = None) -> None:
        self.id_generator = id_generator

    def transform_fields(
        self,
        fields_in: Sequence[FieldIn],
    ) -> Sequence[FieldOut]:
        return [
            FieldOut(
                name=EXTRA_JSON_RPC_REQUEST_ID,
                dest=FieldDestination.EXTRA,
                type_hint=str,
            ),
        ]

    def _new_id(self) -> str:
        if not self.id_generator:
            return str(uuid4())
        return self.id_generator()

    def transform_request(
        self,
        request: HttpRequest,
        fields_in: Sequence[FieldIn],
        fields_out: Sequence[FieldOut],
        data: dict[str, Any],
    ) -> HttpRequest:
        request.extras.append((EXTRA_JSON_RPC_REQUEST_ID, self._new_id()))
        return request

    def __repr__(self):
        return f"{self.__class__.__name__}({self.id_generator})"


class JsonRPCMethod(RequestTransformer):
    def __init__(self, method: str):
        self.method = method

    def transform_fields(
        self,
        fields_in: Sequence[FieldIn],
    ) -> Sequence[FieldOut]:
        return [
            FieldOut(
                name=EXTRA_JSON_RPC_METHOD,
                dest=FieldDestination.EXTRA,
                type_hint=str,
            ),
        ]

    def transform_request(
        self,
        request: HttpRequest,
        fields_in: Sequence[FieldIn],
        fields_out: Sequence[FieldOut],
        data: dict[str, Any],
    ) -> HttpRequest:
        request.extras.append((EXTRA_JSON_RPC_METHOD, self.method))
        return request

    def __repr__(self):
        return f"{self.__class__.__name__}({self.method!r})"


class PackJsonRPC(RequestTransformer):
    def transform_fields(
        self,
        fields_in: Sequence[FieldIn],
    ) -> Sequence[FieldOut]:
        return []

    def transform_request(
        self,
        request: HttpRequest,
        fields_in: Sequence[FieldIn],
        fields_out: Sequence[FieldOut],
        data: dict[str, Any],
    ) -> HttpRequest:
        request_id = get_extra(request, EXTRA_JSON_RPC_REQUEST_ID)
        method = get_extra(request, EXTRA_JSON_RPC_METHOD)
        request.body = {
            "jsonrpc": "2.0",
            "id": request_id,
            "method": method,
            "params": request.body,
        }
        return request

    def __repr__(self):
        return f"{self.__class__.__name__}()"


class UnpackJsonRPC(ResponseTransformer):
    def need_response_body(self, response: HttpResponse) -> bool:
        return True

    def transform_response(
        self,
        request: HttpRequest,
        response: HttpResponse,
    ) -> HttpResponse:
        response.body = response.body.get("result")
        return response

    def __repr__(self):
        return f"{self.__class__.__name__}()"


class JsonRPCErrorRaiser(ResponseTransformer):
    def need_response_body(self, response: HttpResponse) -> bool:
        return True

    def transform_response(
        self,
        request: HttpRequest,
        response: HttpResponse,
    ) -> HttpResponse:
        request_id = get_extra(request, EXTRA_JSON_RPC_REQUEST_ID)
        if request_id != response.body.get("id"):
            raise JsonRPCIdMismatchError
        if error := response.body.get("error"):
            raise JsonRPCError(
                error["code"],
                error["message"],
                error.get("data"),
            )
        return response

    def __repr__(self):
        return f"{self.__class__.__name__}()"


class BuilderParams(TypedDict, total=False):
    http_method: str
    url: UrlSrc
    id_generator: IdGenerator

    request_body_dumper: Dumper | None
    request_body_post_dump: RequestTransformer | None

    response_body_loader: Loader | None
    response_body_pre_load: ResponseTransformer | None
    error_raiser: ResponseTransformer | None
    json_rpc_error_raiser: ResponseTransformer | None


class JsonRPCBuilder:
    def __init__(
        self,
        *transformers: Transformer,
        **params: Unpack[BuilderParams],
    ) -> None:
        self.transformers = transformers
        self.params = params

    def with_params(
        self,
        *transformers: Transformer,
        **params: Unpack[BuilderParams],
    ) -> "JsonRPCBuilder":
        return JsonRPCBuilder(
            *transformers,
            *self.transformers,
            **(self.params | params),
        )

    def _add_request_transformer(
        self,
        spec: MethodSpec,
        transformer: RequestTransformer,
    ):
        spec.request_transformers.append(transformer)
        spec.fields_out.extend(transformer.transform_fields(spec.fields_in))

    def _get_body_field(self, spec: MethodSpec) -> FieldOut | None:
        for field in spec.fields_out:
            if field.dest is FieldDestination.BODY:
                return field
        return None

    def _add_body_transformer(self, spec: MethodSpec):
        body_out = self._get_body_field(spec)
        if body_out is None:
            body_name = None
        else:
            body_name = body_out.name
        for field in spec.fields_in:
            if field.consumed_by:
                continue
            if body_name:
                raise MultipleBodyError(body_name, field.name)
            body_name = field.name
            self._add_request_transformer(spec, Body(field.name))

    def _add_default_request_body_transformers(self, spec: MethodSpec):
        self._add_body_transformer(spec)

        if self._get_body_field(spec):
            dumper = self.params.get("request_body_dumper")
            if dumper:
                self._add_request_transformer(spec, BodyModelDump(dumper))

        id_generator = self.params.get("id_generator", ...)
        if id_generator is ...:
            self._add_request_transformer(spec, JsonRPCIdGenerator())
        elif id_generator:
            self._add_request_transformer(
                spec,
                JsonRPCIdGenerator(id_generator),
            )

        url_src = self.params.get("url") or ""
        self._add_request_transformer(spec, url_transformer(url_src))

        self._add_request_transformer(spec, PackJsonRPC())

        post_dump = self.params.get("request_body_post_dump", ...)
        if post_dump is ...:
            self._add_request_transformer(spec, JsonDump())
        elif post_dump:
            self._add_request_transformer(spec, post_dump)

        http_method = self.params.get("http_method", ...)
        if http_method is ...:
            self._add_request_transformer(spec, Method("POST"))
        elif http_method:
            self._add_request_transformer(spec, Method(http_method))

    def _add_default_response_transformers(self, spec: MethodSpec) -> None:
        error_raiser = self.params.get("error_raiser", ...)
        if error_raiser is ...:
            spec.response_transformers.append(ErrorRaiser())
        elif error_raiser:
            spec.response_transformers.append(error_raiser)

        pre_loader = self.params.get("response_body_pre_load", ...)
        if pre_loader is ...:
            spec.response_transformers.append(JsonLoad())
        elif pre_loader:
            spec.response_transformers.append(pre_loader)

        error_raiser = self.params.get("json_rpc_error_raiser", ...)
        if error_raiser is ...:
            spec.response_transformers.append(JsonRPCErrorRaiser())
        elif error_raiser:
            spec.response_transformers.append(error_raiser)

        spec.response_transformers.append(UnpackJsonRPC())

        loader = self.params.get("response_body_loader")
        if spec.result_type is HttpResponse:
            spec.response_transformers.append(KeepResponse(need_body=False))
        elif (
            loader
            and spec.result_type is not Any
            and spec.result_type is not object
        ):
            spec.response_transformers.append(
                BodyModelLoad(spec.result_type, loader=loader),
            )

    @overload
    def decorate(
        self,
        func: Callable[
            Concatenate[Any, _MethodParamSpec],
            Awaitable[_MethodResultT],
        ],
    ) -> MethodBinder[_MethodParamSpec, _MethodResultT]: ...

    @overload
    def decorate(
        self,
        func: Callable[Concatenate[Any, _MethodParamSpec], _MethodResultT],
    ) -> MethodBinder[_MethodParamSpec, _MethodResultT]: ...

    def decorate(
        self,
        func: Callable[Concatenate[Any, _MethodParamSpec], _MethodResultT],
    ) -> MethodBinder[_MethodParamSpec, _MethodResultT]:
        spec = make_method_spec(
            func,
            transformers=self.transformers,
            is_in_class=True,
        )
        self._add_default_request_body_transformers(spec)
        self._add_default_response_transformers(spec)
        return MethodBinder(spec)

    @overload
    def __call__(
        self,
        func: Callable[
            Concatenate[Any, _MethodParamSpec],
            Awaitable[_MethodResultT],
        ],
    ) -> MethodBinder[_MethodParamSpec, _MethodResultT]: ...

    @overload
    def __call__(
        self,
        func: Callable[Concatenate[Any, _MethodParamSpec], _MethodResultT],
    ) -> MethodBinder[_MethodParamSpec, _MethodResultT]: ...

    @overload
    def __call__(
        self,
        method: str,
        *transformers: Transformer,
        **params: Unpack[BuilderParams],
    ) -> "JsonRPCBuilder": ...

    def __call__(
        self,
        func_or_method: str | Callable,
        *transformers: Transformer,
        **params: Unpack[BuilderParams],
    ) -> Any:
        if transformers or params:
            instance = self.with_params(*transformers, **params)
        else:
            instance = self
        if isinstance(func_or_method, str):
            return instance.with_params(JsonRPCMethod(func_or_method))
        else:
            return instance.decorate(func_or_method)



from typing import Any

import pytest
import pytest_asyncio
import requests

from descanso import RestBuilder
from descanso.http.requests import RequestsClient
from descanso.jsonrpc import JsonRPCBuilder, JsonRPCError
from .data import req_resp


@pytest_asyncio.fixture
def client(server_addr):
    session = requests.Session()
    yield RequestsClient(server_addr, session)


@pytest.mark.parametrize(("req", "expected_resp"), req_resp())
def test_requests(client, req, expected_resp):
    with client.send_request(req) as resp:
        resp.load_body()
        assert resp == expected_resp


def test_rest_requests(server_addr):
    rest = RestBuilder()

    class Client(RequestsClient):
        @rest.get("/json")
        def do_get(self, body: dict) -> dict: ...

    client = Client(server_addr, requests.Session())
    resp = client.do_get({"x": 1})
    assert resp == {"y": 2}


def test_jsonrpc_requests(server_addr):
    jsonrpc = JsonRPCBuilder(url="jsonrpc")

    class Client(RequestsClient):
        @jsonrpc("good")
        def do_good(self, body: Any) -> Any: ...

        @jsonrpc("bad")
        def do_bad(self) -> Any: ...

        @jsonrpc("invalid")
        def do_invalid(self) -> Any: ...

    client = Client(server_addr, requests.Session())
    resp = client.do_good([42])
    assert resp == 42

    with pytest.raises(JsonRPCError) as e:
        client.do_bad()
    assert e.value.code == -32000
    assert e.value.message == "My error"

    with pytest.raises(JsonRPCError) as e:
        client.do_invalid()
    assert e.value.code == -32601
    assert e.value.message == "Method not found"



from typing import Any

from descanso import Loader
from descanso.request import HttpRequest
from descanso.response import HttpResponse
from descanso.response_transformers import (
    BodyModelLoad,
    JsonLoad,
    KeepResponse,
)


def test_keep_response():
    json_load = KeepResponse(need_body=True)
    response = HttpResponse(
        status_code=200,
        status_text="OK",
        body='{"x": 1}',
    )
    response2 = json_load.transform_response(HttpRequest(), response)
    assert str(json_load)
    assert response2 == HttpResponse(
        status_code=200,
        status_text="OK",
        body=response,
    )


def test_json_load():
    json_load = JsonLoad()
    response = HttpResponse(
        status_code=200,
        status_text="OK",
        body='{"x": 1}',
    )
    response = json_load.transform_response(HttpRequest(), response)
    assert str(json_load)
    assert response == HttpResponse(
        status_code=200,
        status_text="OK",
        body={"x": 1},
    )


class Model:
    pass


class StubLoader(Loader):
    def load(self, data: Any, class_: Any) -> Any:
        return ["stub", data, class_]


def test_model_load():
    model_load = BodyModelLoad(
        type_hint=Model,
        loader=StubLoader(),
    )
    response = HttpResponse(
        status_code=200,
        status_text="OK",
        body="x",
    )
    response = model_load.transform_response(HttpRequest(), response)
    assert str(model_load)
    assert response == HttpResponse(
        status_code=200,
        status_text="OK",
        body=["stub", "x", Model],
    )



import inspect
from collections.abc import Callable, Sequence
from typing import Any, get_type_hints

from .method_spec import MethodSpec
from .request import FieldIn, RequestTransformer
from .response import ResponseTransformer


def get_func_fields(func: Callable, *, is_in_class) -> list[FieldIn]:
    signature = inspect.signature(func)
    hints = get_type_hints(func)
    fields = [
        FieldIn(
            name=arg.name,
            type_hint=hints.get(arg.name, Any),
            consumed_by=[],
        )
        for arg in signature.parameters.values()
    ]
    if is_in_class:
        del fields[0]
    return fields


def get_result_type(func: Callable) -> Any:
    hints = get_type_hints(func)
    return hints.get("return", Any)


def make_method_spec(
    func: Callable,
    *,
    transformers: Sequence[RequestTransformer | ResponseTransformer],
    is_in_class: bool,
):
    fields_in = get_func_fields(func, is_in_class=is_in_class)
    fields_out = []
    request_transformers = [
        r for r in transformers if isinstance(r, RequestTransformer)
    ]
    for r in request_transformers:
        fields_out.extend(r.transform_fields(fields_in))

    return MethodSpec(
        func=func,
        name=func.__name__,
        doc=func.__doc__,
        fields_in=fields_in,
        fields_out=fields_out,
        result_type=get_result_type(func),
        request_transformers=request_transformers,
        response_transformers=[
            r for r in transformers if isinstance(r, ResponseTransformer)
        ],
    )



from inspect import Parameter, signature
from typing import TypedDict, get_args, get_origin, get_type_hints

try:
    from typing import Unpack
except ImportError:
    Unpack = None


class _ExampleTypedDict(TypedDict):
    pass


TypedDictMeta = type(_ExampleTypedDict)


def unpack_signature(app, obj, bound_method):
    if not bound_method:
        return

    sig = signature(obj)
    hints = get_type_hints(obj)
    new_params = []
    for param in sig.parameters.values():
        hint = hints.get(param.name)
        # Not an Unpack[...] annotation → keep param as-is
        if get_origin(hint) != Unpack:
            new_params.append(param)
            continue
        # Expect Unpack[SomeTypedDict]
        args = get_args(hint)
        if len(args) != 1:
            new_params.append(param)
            continue
        typed_dict_type = args[0]
        if type(typed_dict_type) != TypedDictMeta:
            new_params.append(param)
            continue
        for key, annotation in get_type_hints(typed_dict_type).items():
            new_params.append(
                Parameter(
                    name=key,
                    kind=Parameter.KEYWORD_ONLY,
                    annotation=annotation,
                ),
            )

    obj.__signature__ = sig.replace(parameters=new_params)


def setup(app):
    if Unpack:
        app.connect("autodoc-before-process-signature", unpack_signature)



import urllib.parse
from collections.abc import Iterator, Sequence
from contextlib import contextmanager

from kiss_headers import parse_it
from requests import Response, Session

from descanso.client import (
    SyncClient,
    SyncResponseWrapper,
)
from descanso.request import HttpRequest, RequestTransformer
from descanso.response import ResponseTransformer
from descanso.utils import ensure_trailing_slash


class RequestsResponseWrapper(SyncResponseWrapper):
    def __init__(self, response: Response) -> None:
        self.status_code = response.status_code
        self.status_text = response.reason
        self.body = None
        self.headers = parse_it(response.headers)
        self._raw_response = response

    def load_body(self) -> None:
        self.body = self._raw_response.content


class RequestsClient(SyncClient):
    def __init__(
        self,
        base_url: str,
        session: Session,
        transformers: Sequence[RequestTransformer | ResponseTransformer] = (),
    ) -> None:
        super().__init__(
            transformers=transformers,
        )
        self._base_url = ensure_trailing_slash(base_url)
        self._session = session

    @contextmanager
    def send_request(
        self,
        request: HttpRequest,
    ) -> Iterator[SyncResponseWrapper]:
        params = [(k, v) for k, v in request.query_params if v is not None]
        resp = self._session.request(
            method=request.method,
            url=urllib.parse.urljoin(self._base_url, request.url),
            headers=request.headers,
            data=request.body,
            params=params,
            files=[
                (name, (data.filename, data.contents, data.content_type))
                for name, data in request.files
            ],
        )
        yield RequestsResponseWrapper(resp)



from collections.abc import Sequence
from contextlib import AbstractAsyncContextManager, AbstractContextManager
from typing import Any, Protocol

from descanso.request import HttpRequest, RequestTransformer
from descanso.response import HttpResponse, ResponseTransformer


class Dumper(Protocol):
    def dump(self, data: Any, class_: Any) -> Any:
        raise NotImplementedError


class Loader(Protocol):
    def load(self, data: Any, class_: Any) -> Any:
        raise NotImplementedError


class BaseClient:
    def __init__(
        self,
        transformers: Sequence[RequestTransformer | ResponseTransformer],
    ):
        self.request_transformers = [
            r for r in transformers if isinstance(r, RequestTransformer)
        ]
        self.response_transformers = [
            r for r in transformers if isinstance(r, ResponseTransformer)
        ]


class SyncResponseWrapper(HttpResponse):
    def load_body(self) -> None:
        raise NotImplementedError


class SyncClient(BaseClient):
    def send_request(
        self,
        request: HttpRequest,
    ) -> AbstractContextManager[SyncResponseWrapper]:
        raise NotImplementedError


class AsyncResponseWrapper(HttpResponse):
    async def aload_body(self) -> None:
        raise NotImplementedError


class AsyncClient(BaseClient):
    def asend_request(
        self,
        request: HttpRequest,
    ) -> AbstractAsyncContextManager[AsyncResponseWrapper]:
        raise NotImplementedError



from inspect import getcallargs
from typing import (
    Any,
)

from .client import (
    AsyncClient,
    AsyncResponseWrapper,
    BaseClient,
    SyncClient,
    SyncResponseWrapper,
)
from .method_spec import MethodSpec
from .request import HttpRequest
from .response import HttpResponse


def make_request(
    client: BaseClient,
    spec: MethodSpec,
    args: dict[str, Any],
) -> HttpRequest:
    request = HttpRequest()
    for transformer in spec.request_transformers:
        transformer.transform_request(
            request,
            spec.fields_in,
            spec.fields_out,
            args,
        )
    for transformer in client.request_transformers:
        transformer.transform_request(
            request,
            spec.fields_in,
            spec.fields_out,
            args,
        )
    return request


def make_response_sync(
    client: BaseClient,
    spec: MethodSpec,
    request: HttpRequest,
    response: SyncResponseWrapper,
) -> Any:
    loaded = False
    for transformer in spec.response_transformers:
        if not loaded and transformer.need_response_body(response):
            response.load_body()
            loaded = True
        response = transformer.transform_response(request, response)
    for transformer in client.response_transformers:
        if not loaded and transformer.need_response_body(response):
            response.load_body()
            loaded = True
        transformer.transform_response(request, response)
    return response.body


async def make_response_async(
    client: BaseClient,
    spec: MethodSpec,
    request: HttpRequest,
    response: AsyncResponseWrapper,
) -> Any:
    loaded = False
    for transformer in spec.response_transformers:
        if not loaded and transformer.need_response_body(response):
            await response.aload_body()
            loaded = True
        response = transformer.transform_response(request, response)
    for transformer in client.response_transformers:
        if not loaded and transformer.need_response_body(response):
            await response.aload_body()
            loaded = True
        transformer.transform_response(request, response)
    return response.body


def need_response_body(
    spec: MethodSpec,
    response: HttpResponse,
) -> bool:
    for transformer in spec.response_transformers:
        if transformer.need_response_body(response):
            return True
    return False


class BoundSyncMethod:
    __slots__ = ("_client", "_spec")

    def __init__(self, spec: MethodSpec, client: SyncClient) -> None:
        self._spec = spec
        self._client = client

    def __call__(self, *args, **kwargs):
        args = getcallargs(self._spec.func, self._client, *args, **kwargs)
        request = make_request(self._client, self._spec, args)
        with self._client.send_request(request) as response:
            return make_response_sync(
                self._client,
                self._spec,
                request,
                response,
            )


class BoundAsyncMethod:
    __slots__ = ("_client", "_spec")

    def __init__(self, spec: MethodSpec, client: AsyncClient) -> None:
        self._spec = spec
        self._client = client

    async def __call__(self, *args, **kwargs):
        args = getcallargs(self._spec.func, self._client, *args, **kwargs)
        request = make_request(self._client, self._spec, args)
        async with self._client.asend_request(request) as response:
            return await make_response_async(
                self._client,
                self._spec,
                request,
                response,
            )



import pytest

from descanso.request import FieldIn, HttpRequest
from descanso.request_transformers import (
    DeepObjectQuery,
    DelimiterQuery,
    FormQuery,
    PhpStyleQuery,
)
from tests.request_transformers.utills import consumed_fields


@pytest.fixture
def fields_in():
    return [FieldIn("x", int)]


@pytest.mark.parametrize(
    ("transformer", "expected_params"),
    [
        (DeepObjectQuery(), [("x[]", "1"), ("x[]", "2")]),
        (FormQuery(), [("x", "1"), ("x", "2")]),
        (DelimiterQuery(), [("x", "1,2")]),
        (DelimiterQuery("|"), [("x", "1|2")]),
    ],
)
def test_list(transformer, expected_params, fields_in):
    assert str(transformer)
    assert transformer.transform_fields(fields_in) == []
    assert consumed_fields(fields_in, transformer) == []
    request = HttpRequest(
        query_params=[
            ("x", [1, 2]),
        ],
    )
    request = transformer.transform_request(request, fields_in, [], {"x": 1})
    assert request == HttpRequest(query_params=expected_params)


@pytest.mark.parametrize(
    ("transformer", "expected_params"),
    [
        (DeepObjectQuery(), [("x[a]", "1"), ("x[b]", "2")]),
        (FormQuery(), [("a", "1"), ("b", "2")]),
        (DelimiterQuery(), [("x", "a,1,b,2")]),
        (DelimiterQuery("|"), [("x", "a|1|b|2")]),
    ],
)
def test_dict(transformer, expected_params, fields_in):
    assert str(transformer)
    assert transformer.transform_fields(fields_in) == []
    assert consumed_fields(fields_in, transformer) == []
    request = HttpRequest(
        query_params=[
            ("x", {"a": 1, "b": 2, "c": None}),
        ],
    )
    request = transformer.transform_request(request, fields_in, [], {"x": 1})
    assert request == HttpRequest(query_params=expected_params)


def test_nested_php_style(fields_in):
    transformer = PhpStyleQuery()
    assert str(transformer)
    assert transformer.transform_fields(fields_in) == []
    assert consumed_fields(fields_in, transformer) == []
    request = HttpRequest(
        query_params=[
            (
                "x",
                {
                    "a": 1,
                    "b": [
                        {"a": 2, "b": 3},
                        {"a": 4, "b": 5},
                    ],
                    "c": None,
                },
            ),
        ],
    )
    expected_params = [
        ("x[a]", "1"),
        ("x[b][0][a]", "2"),
        ("x[b][0][b]", "3"),
        ("x[b][1][a]", "4"),
        ("x[b][1][b]", "5"),
    ]
    request = transformer.transform_request(request, fields_in, [], {"x": 1})
    assert request == HttpRequest(query_params=expected_params)



import pytest

from descanso.exceptions import ClientError, HttpStatusError, ServerError
from descanso.request import HttpRequest
from descanso.response import HttpResponse
from descanso.response_transformers import ErrorRaiser


@pytest.mark.parametrize(
    ("status_code", "status_text", "except_codes", "codes"),
    [
        (200, "ok", None, None),
        (201, "created", [201], None),
        (
            204,
            "no_content",
            [200, 201, 202, 204],
            [202, 201, 200],
        ),
        (
            201,
            "created",
            [200, 201, 202, 204],
            [202, 204],
        ),
    ],
)
def test_ok(
    status_code: int,
    status_text: str,
    except_codes: list[int],
    codes: list[int],
):
    error_raiser = ErrorRaiser(except_codes=except_codes, codes=codes)
    response = HttpResponse(status_code=status_code, status_text=status_text)

    transformed_response = error_raiser.transform_response(
        HttpRequest(),
        response,
    )

    assert str(error_raiser)
    assert transformed_response == response


@pytest.mark.parametrize(
    ("status_code", "status_text", "except_codes", "codes", "error_type"),
    [
        # Client Error
        (404, "not_found", None, None, ClientError),
        (400, "bad_request", [], [], ClientError),
        (201, "created", None, [201], ClientError),
        (200, "ok", [201], None, ClientError),
        (302, "found", [200, 201, 202, 204], [202, 204], ClientError),
        (202, "accepted", [202, 204], [200, 201, 202, 204], ClientError),
        # Server Error
        (500, "internal_server_error", None, None, ServerError),
        (502, "bad_gateway", None, None, ServerError),
    ],
)
def test_error(
    status_code: int,
    status_text: str,
    except_codes: list[int],
    codes: list[int],
    error_type: type[HttpStatusError],
):
    error_raiser = ErrorRaiser(codes=codes, except_codes=except_codes)
    response = HttpResponse(status_code=status_code, status_text=status_text)

    with pytest.raises(error_type) as exc_info:
        error_raiser.transform_response(HttpRequest(), response)

    exc = exc_info.value
    assert str(error_raiser)
    assert isinstance(exc, error_type)
    assert exc.status_code == response.status_code
    assert exc.status_text == response.status_text



from collections.abc import Callable
from dataclasses import dataclass
from typing import Any, Generic, ParamSpec, TypeVar

from .request import FieldIn, FieldOut, RequestTransformer
from .response import ResponseTransformer

_MethodResultT = TypeVar("_MethodResultT")
_MethodParamSpec = ParamSpec("_MethodParamSpec")


@dataclass
class MethodSpec(Generic[_MethodParamSpec, _MethodResultT]):
    name: str
    doc: str
    fields_in: list[FieldIn]
    fields_out: list[FieldOut]
    result_type: Any
    func: Callable[_MethodParamSpec, _MethodResultT]
    request_transformers: list[RequestTransformer]
    response_transformers: list[ResponseTransformer]



from descanso import JsonRPCBuilder
from descanso.jsonrpc import (
    JsonRPCErrorRaiser,
    JsonRPCIdGenerator,
    JsonRPCMethod,
    PackJsonRPC,
    UnpackJsonRPC,
)
from descanso.request_transformers import (
    Body,
    BodyModelDump,
    JsonDump,
    Method,
    Skip,
    Url,
)
from descanso.response_transformers import BodyModelLoad, ErrorRaiser, JsonLoad
from .test_rest import StubConverter
from .utils import dirty


class Model:
    pass


def test_url():
    jsonrpc = JsonRPCBuilder(url="/foo")

    class Api:
        @jsonrpc("methodname")
        def do(self, body: int) -> Model:
            """Hello"""

    assert Api.do.spec.name == "do"
    assert Api.do.spec.doc == "Hello"
    assert Api.do.spec.request_transformers == [
        dirty[JsonRPCMethod](method="methodname"),
        dirty[Body](arg="body"),
        dirty[JsonRPCIdGenerator](id_generator=None),
        dirty[Url](original_template="/foo"),
        dirty[PackJsonRPC](),
        dirty[JsonDump](),
        dirty[Method](method="POST"),
    ]
    assert Api.do.spec.response_transformers == [
        dirty[ErrorRaiser](),
        dirty[JsonLoad](),
        dirty[JsonRPCErrorRaiser](),
        dirty[UnpackJsonRPC](),
    ]


def id_generator():
    return ""


def test_params():
    req_additional = Skip("0")
    resp_additional = ErrorRaiser(codes=[200])
    request_body_dumper = StubConverter()
    response_body_loader = StubConverter()
    request_body_post_dump = Skip("2")
    error_raiser = ErrorRaiser(codes=[400])
    response_body_pre_load = ErrorRaiser(codes=[500])
    json_rpc_error_raiser = JsonRPCErrorRaiser()

    jsonrpc = JsonRPCBuilder(
        req_additional,
        resp_additional,
        http_method="GET",
        id_generator=id_generator,
        request_body_dumper=request_body_dumper,
        response_body_loader=response_body_loader,
        request_body_post_dump=request_body_post_dump,
        error_raiser=error_raiser,
        json_rpc_error_raiser=json_rpc_error_raiser,
        response_body_pre_load=response_body_pre_load,
    )

    class Api:
        @jsonrpc("methodname")
        def do(self, data: int) -> Model:
            """Hello"""

    assert Api.do.spec.name == "do"
    assert Api.do.spec.doc == "Hello"
    assert Api.do.spec.request_transformers == [
        dirty[JsonRPCMethod](method="methodname"),
        req_additional,
        dirty[Body](arg="data"),
        dirty[BodyModelDump](dumper=request_body_dumper),
        dirty[JsonRPCIdGenerator](id_generator=id_generator),
        dirty[Url](original_template=""),
        dirty[PackJsonRPC](),
        request_body_post_dump,
        dirty[Method](method="GET"),
    ]
    assert Api.do.spec.response_transformers == [
        resp_additional,
        error_raiser,
        response_body_pre_load,
        json_rpc_error_raiser,
        dirty[UnpackJsonRPC](),
        dirty[BodyModelLoad](type_hint=Model, loader=response_body_loader),
    ]


def test_override():
    jsonrpc = JsonRPCBuilder(url="/foo")

    class Api:
        @jsonrpc("methodname", url="/bar")
        def do(self, data: int) -> Model:
            """Hello"""

    assert Api.do.spec.name == "do"
    assert Api.do.spec.doc == "Hello"
    assert Api.do.spec.request_transformers == [
        dirty[JsonRPCMethod](method="methodname"),
        dirty[Body](arg="data"),
        dirty[JsonRPCIdGenerator](id_generator=None),
        dirty[Url](original_template="/bar"),
        dirty[PackJsonRPC](),
        dirty[JsonDump](),
        dirty[Method](method="POST"),
    ]
    assert Api.do.spec.response_transformers == [
        dirty[ErrorRaiser](),
        dirty[JsonLoad](),
        dirty[JsonRPCErrorRaiser](),
        dirty[UnpackJsonRPC](),
    ]



import itertools
import json
import string
from collections.abc import Callable, Iterator, Sequence
from inspect import getfullargspec
from typing import Any, get_type_hints

from .client import Dumper
from .request import (
    BaseRequestTransformer,
    FieldDestination,
    FieldIn,
    FieldOut,
    FileData,
    HttpRequest,
    KeyValue,
)


def _base_field_name(field_name: str) -> str:
    field_name, *_ = field_name.split(".", 1)
    field_name, *_ = field_name.split("[", 1)
    return field_name


def get_params_from_string(template: str) -> list[str]:
    parsed_format = list(string.Formatter().parse(template))
    return [_base_field_name(x[1]) for x in parsed_format if x[1]]


def get_params_from_callable(
    template: Callable[..., Any],
) -> list[str]:
    url_template_func_arg_spec = getfullargspec(template)
    return url_template_func_arg_spec.args


DataTemplate = Callable[..., Any] | str | None


class DestTransformer(BaseRequestTransformer):
    def __init__(
        self,
        name_out: str,
        template: DataTemplate,
        dest: FieldDestination,
    ) -> None:
        self.name_out = name_out
        self.dest = dest
        self.original_template = template
        self.type_hint = Any
        if template is None:
            self.template = lambda **kwargs: kwargs[name_out]
            self.args = [name_out]
            self.type_hint = Any
        elif isinstance(template, str):
            self.template = template.format
            self.args = get_params_from_string(template)
            self.type_hint = str
        else:
            self.template = template
            self.args = get_params_from_callable(template)
            self.type_hint = get_type_hints(template).get("return", Any)

    def transform_fields(
        self,
        fields: Sequence[FieldIn],
    ) -> Sequence[FieldOut]:
        type_hint = self.type_hint
        for field in fields:
            if field.name in self.args:
                field.consumed_by.append(self)
                if self.original_template is None:
                    type_hint = field.type_hint
        return [
            FieldOut(
                name=self.name_out,
                dest=self.dest,
                type_hint=type_hint,
            ),
        ]

    def transform_request(
        self,
        request: HttpRequest,
        fields_in: Sequence[FieldIn],
        fields_out: Sequence[FieldOut],
        data: dict[str, Any],
    ) -> HttpRequest:
        request_field = getattr(request, self.dest.value)
        data = self.template(
            **{k: v for k, v in data.items() if k in self.args},
        )
        request_field.append((self.name_out, data))
        return request

    def __repr__(self):
        return (
            f"{self.__class__.__name__}("
            f"{self.name_out!r}, "
            f"{self.original_template!r}, "
            f"{self.dest!r}"
            f")"
        )


class Header(DestTransformer):
    def __init__(self, header: str, template: DataTemplate = None):
        super().__init__(
            name_out=header,
            template=template,
            dest=FieldDestination.HEADER,
        )

    def transform_request(
        self,
        request: HttpRequest,
        fields_in: Sequence[FieldIn],
        fields_out: Sequence[FieldOut],
        data: dict[str, Any],
    ) -> HttpRequest:
        data = self.template(
            **{k: v for k, v in data.items() if k in self.args},
        )
        request.headers[self.name_out] = str(data)
        return request


class Extra(DestTransformer):
    def __init__(self, header: str, template: DataTemplate = None):
        super().__init__(
            name_out=header,
            template=template,
            dest=FieldDestination.EXTRA,
        )


class Query(DestTransformer):
    def __init__(self, name_out: str, template: DataTemplate = None):
        super().__init__(
            name_out=name_out,
            template=template,
            dest=FieldDestination.QUERY,
        )


class Url(BaseRequestTransformer):
    def __init__(self, template: Callable | str):
        self._field_out = FieldOut(
            name=None,
            dest=FieldDestination.URL,
            type_hint=str,
        )
        self.original_template = template
        if isinstance(template, str):
            self.template = template.format
            self.args = get_params_from_string(template)
        else:
            self.template = template
            self.args = get_params_from_callable(template)

    def transform_fields(
        self,
        fields: Sequence[FieldIn],
    ) -> Sequence[FieldOut]:
        for field in fields:
            if field.name in self.args:
                field.consumed_by.append(self)
        return [self._field_out]

    def transform_request(
        self,
        request: HttpRequest,
        fields_in: Sequence[FieldIn],
        fields_out: Sequence[FieldOut],
        data: dict[str, Any],
    ) -> HttpRequest:
        request.url = self.template(
            **{k: v for k, v in data.items() if k in self.args},
        )
        return request

    def __repr__(self):
        return f"{self.__class__.__name__}({self.original_template!r})"


class File(BaseRequestTransformer):
    def __init__(
        self,
        arg: str,
        filefield: str | None = None,
        filename: str | None = None,
        content_type: str | None = None,
    ):
        self.arg = arg
        self.filefield = filefield or arg
        self.filename = filename
        self.content_type = content_type

    def transform_fields(
        self,
        fields: Sequence[FieldIn],
    ) -> Sequence[FieldOut]:
        for field in fields:
            if field.name == self.arg:
                field.consumed_by.append(self)
                return [
                    FieldOut(
                        name=self.filefield,
                        dest=FieldDestination.FILE,
                        type_hint=field.type_hint,
                    ),
                ]
        return []

    def transform_request(
        self,
        request: HttpRequest,
        fields_in: Sequence[FieldIn],
        fields_out: Sequence[FieldOut],
        data: dict[str, Any],
    ) -> HttpRequest:
        request.files.append(
            (
                self.filefield,
                FileData(
                    filename=self.filename,
                    contents=data[self.arg],
                    content_type=self.content_type,
                ),
            ),
        )
        return request

    def __repr__(self):
        return (
            f"{self.__class__.__name__}("
            f"{self.arg!r}, "
            f"{self.filefield!r}, "
            f"{self.filename!r}, "
            f"{self.content_type!r}"
            f")"
        )


class Body(BaseRequestTransformer):
    def __init__(self, arg: str):
        self.arg = arg

    def transform_fields(
        self,
        fields: Sequence[FieldIn],
    ) -> Sequence[FieldOut]:
        for field in fields:
            if field.name == self.arg:
                field.consumed_by.append(self)
                return [
                    FieldOut(
                        name=None,
                        dest=FieldDestination.BODY,
                        type_hint=field.type_hint,
                    ),
                ]
        return []

    def transform_request(
        self,
        request: HttpRequest,
        fields_in: Sequence[FieldIn],
        fields_out: Sequence[FieldOut],
        data: dict[str, Any],
    ) -> HttpRequest:
        if self.arg not in data:
            return request
        request.body = data[self.arg]
        return request

    def __repr__(self):
        return f"{self.__class__.__name__}({self.arg!r})"


class BodyModelDump(BaseRequestTransformer):
    def __init__(self, dumper: Dumper) -> None:
        self.dumper = dumper

    def transform_request(
        self,
        request: HttpRequest,
        fields_in: Sequence[FieldIn],
        fields_out: Sequence[FieldOut],
        data: dict[str, Any],
    ) -> HttpRequest:
        type_hint = next(
            (
                f.type_hint
                for f in fields_out
                if f.dest == FieldDestination.BODY
            ),
            Any,
        )
        request.body = self.dumper.dump(request.body, type_hint)
        return request

    def __repr__(self):
        return f"{self.__class__.__name__}({self.dumper!r})"


class QueryModelDump(BaseRequestTransformer):
    def __init__(self, dumper: Dumper) -> None:
        self.dumper = dumper

    def transform_request(
        self,
        request: HttpRequest,
        fields_in: Sequence[FieldIn],
        fields_out: Sequence[FieldOut],
        data: dict[str, Any],
    ) -> HttpRequest:
        types = {
            f.name: f.type_hint
            for f in fields_out
            if f.dest == FieldDestination.QUERY
        }
        request.query_params = [
            (name, self.dumper.dump(value, types.get(name, Any)))
            for name, value in request.query_params
        ]
        return request

    def __repr__(self):
        return f"{self.__class__.__name__}({self.dumper!r})"


class JsonDump(BaseRequestTransformer):
    def transform_request(
        self,
        request: HttpRequest,
        fields_in: Sequence[FieldIn],
        fields_out: Sequence[FieldOut],
        data: dict[str, Any],
    ) -> HttpRequest:
        request.body = json.dumps(request.body)
        request.headers["Content-Type"] = "application/json"
        return request

    def __repr__(self):
        return f"{self.__class__.__name__}()"


class Method(BaseRequestTransformer):
    def __init__(self, method: str) -> None:
        self.method = method

    def transform_request(
        self,
        request: HttpRequest,
        fields_in: Sequence[FieldIn],
        fields_out: Sequence[FieldOut],
        data: dict[str, Any],
    ) -> HttpRequest:
        request.method = self.method
        return request

    def __repr__(self):
        return f"{self.__class__.__name__}({self.method!r})"


class Skip(BaseRequestTransformer):
    def __init__(
        self,
        arg: str | None = None,
    ):
        self.arg = arg

    def transform_fields(
        self,
        fields: Sequence[FieldIn],
    ) -> Sequence[FieldOut]:
        if not self.arg:
            return []
        for field in fields:
            if field.name == self.arg:
                field.consumed_by.append(self)
        return []

    def __repr__(self):
        return f"{self.__class__.__name__}({self.arg!r})"


class DelimiterQuery(BaseRequestTransformer):
    def __init__(self, separator: str = ",") -> None:
        self.separator = separator

    def transform_request(
        self,
        request: HttpRequest,
        fields_in: Sequence[FieldIn],
        fields_out: Sequence[FieldOut],
        data: dict[str, Any],
    ) -> HttpRequest:
        for i, (name, value) in enumerate(request.query_params):
            if value is None:
                continue
            if isinstance(value, list):
                request.query_params[i] = (
                    name,
                    self.separator.join(map(str, value)),
                )
            elif isinstance(value, dict):
                request.query_params[i] = (
                    name,
                    self.separator.join(
                        itertools.chain.from_iterable(
                            (key, str(single_value))
                            for key, single_value in value.items()
                            if single_value is not None
                        ),
                    ),
                )
        return request

    def __repr__(self):
        return f"{self.__class__.__name__}({self.separator!r})"


class DeepObjectQuery(BaseRequestTransformer):
    def transform_request(
        self,
        request: HttpRequest,
        fields_in: Sequence[FieldIn],
        fields_out: Sequence[FieldOut],
        data: dict[str, Any],
    ) -> HttpRequest:
        new_params = []
        for name, value in request.query_params:
            new_params.extend(self._dump(name, value))
        request.query_params = new_params
        return request

    def _dump(self, name: str, value: Any) -> Iterator[KeyValue[str]]:
        if value is None:
            return
        if isinstance(value, list):
            for single_value in value:
                yield f"{name}[]", str(single_value)
        elif isinstance(value, dict):
            for key, single_value in value.items():
                if single_value is None:
                    continue
                yield f"{name}[{key}]", str(single_value)
        else:
            yield name, str(value)

    def __repr__(self):
        return f"{self.__class__.__name__}()"


class PhpStyleQuery(BaseRequestTransformer):
    def transform_request(
        self,
        request: HttpRequest,
        fields_in: Sequence[FieldIn],
        fields_out: Sequence[FieldOut],
        data: dict[str, Any],
    ) -> HttpRequest:
        new_params = []
        for name, value in request.query_params:
            new_params.extend(self._dump(name, value))
        request.query_params = new_params
        return request

    def _dump(self, prefix: str, data: Any) -> Iterator[KeyValue[str]]:
        if data is None:
            return
        elif isinstance(data, list):
            for i, value in enumerate(data):
                yield from self._dump(prefix + f"[{i}]", value)
        elif isinstance(data, dict):
            for key, value in data.items():
                yield from self._dump(prefix + f"[{key}]", value)
        else:
            yield prefix, str(data)

    def __repr__(self):
        return f"{self.__class__.__name__}()"


class FormQuery(BaseRequestTransformer):
    def transform_request(
        self,
        request: HttpRequest,
        fields_in: Sequence[FieldIn],
        fields_out: Sequence[FieldOut],
        data: dict[str, Any],
    ) -> HttpRequest:
        new_params = []
        for name, value in request.query_params:
            new_params.extend(self._dump(name, value))
        request.query_params = new_params
        return request

    def _dump(self, name: str, value: Any) -> Iterator[KeyValue[str]]:
        if value is None:
            return
        if isinstance(value, list):
            for single_value in value:
                yield f"{name}", str(single_value)
        elif isinstance(value, dict):
            for key, single_value in value.items():
                if single_value is None:
                    continue
                yield f"{key}", str(single_value)
        else:
            yield name, str(value)

    def __repr__(self):
        return f"{self.__class__.__name__}()"



from io import BytesIO

from dirty_equals import IsPartialDataclass
from kiss_headers import Header, Headers

from descanso.request import FileData, HttpRequest
from descanso.response import HttpResponse


class HasHeaders:
    def __init__(self, *headers: Header) -> None:
        self.orig = headers
        self.headers = headers
        self.headers: dict[str, Header | list[Header]] = {}
        for header in headers:
            if header.name in self.headers:
                old = self.headers[header.name]
                if isinstance(old, list):
                    old.append(header)
                else:
                    self.headers[header.name] = [old, header]
            else:
                self.headers[header.name] = header

    def __eq__(self, other: HttpResponse):
        for name, value in self.headers.items():
            assert other.headers.get(name) == value
        return True


def req_resp():
    return [
        (
            HttpRequest(
                url="/query_xxy",
                query_params=[
                    ("x", 1),
                    ("x", 2),
                    ("y", 3),
                ],
            ),
            IsPartialDataclass(
                status_code=200,
                body=b"ok",
            ),
        ),
        (
            HttpRequest(
                url="/conflict",
            ),
            IsPartialDataclass(
                status_code=409,
                status_text="test_conflict",
                body=b"test_conflict_body",
            ),
        ),
        (
            HttpRequest(
                url="/delete",
                method="DELETE",
            ),
            IsPartialDataclass(
                status_code=204,
            ),
        ),
        (
            HttpRequest(
                url="/headers",
                method="GET",
                headers=Headers(
                    Header("X-Test", "1"),
                ),
            ),
            HasHeaders(
                Header("X-Test", "2"),
                Header("X-Test", "3"),
                Header("Another", "value"),
            ),
        ),
        (
            HttpRequest(
                url="/files",
                files=[
                    ("x1", FileData(BytesIO(b"1"), filename="x1")),
                    ("x2", FileData(b"2", filename="x2")),
                    (
                        "x3",
                        FileData(
                            b"3",
                            filename="z3",
                            content_type="test_content_type",
                        ),
                    ),
                ],
                method="POST",
            ),
            IsPartialDataclass(
                status_code=200,
            ),
        ),
        (
            HttpRequest(
                url="/form",
                method="POST",
                body={
                    "field1": "1",
                    "field2": "2",
                },
            ),
            IsPartialDataclass(
                status_code=200,
            ),
        ),
    ]



from typing import Any

import pytest
from kiss_headers import Header, Headers

from descanso import Dumper
from descanso.request import (
    FieldDestination,
    FieldIn,
    FieldOut,
    HttpRequest,
    RequestTransformer,
)
from descanso.request_transformers import (
    BodyModelDump,
    JsonDump,
    QueryModelDump,
)
from tests.request_transformers.utills import consumed_fields


class QueryModel:
    pass


class BodyModel:
    pass


class StubDumper(Dumper):
    def dump(self, data: Any, class_: Any) -> Any:
        return ["stub", data, class_]


@pytest.fixture
def fields_in():
    return [FieldIn("x", int)]


@pytest.fixture
def fields_out():
    return [
        FieldOut(name="x", dest=FieldDestination.QUERY, type_hint=QueryModel),
        FieldOut(name=None, dest=FieldDestination.BODY, type_hint=BodyModel),
    ]


@pytest.fixture
def model_request():
    return HttpRequest(
        query_params=[
            ("x", "x_value"),
            ("y", "y_value"),
        ],
        body="body_value",
    )


@pytest.mark.parametrize(
    ("transformer", "expected_request"),
    [
        (
            QueryModelDump(StubDumper()),
            HttpRequest(
                query_params=[
                    ("x", ["stub", "x_value", QueryModel]),
                    ("y", ["stub", "y_value", Any]),
                ],
                body="body_value",
            ),
        ),
        (
            BodyModelDump(StubDumper()),
            HttpRequest(
                query_params=[
                    ("x", "x_value"),
                    ("y", "y_value"),
                ],
                body=["stub", "body_value", BodyModel],
            ),
        ),
    ],
)
def test_body_model_dump(
    fields_in,
    fields_out,
    model_request,
    transformer: RequestTransformer,
    expected_request: HttpRequest,
):
    assert str(transformer)
    assert transformer.transform_fields(fields_in) == []
    assert consumed_fields(fields_in, transformer) == []
    request = transformer.transform_request(
        model_request,
        fields_in,
        fields_out,
        {"x": 0, "y": 0},
    )
    assert request == expected_request


def test_body_json_dump(fields_in, fields_out):
    transformer = JsonDump()
    assert str(transformer)
    assert transformer.transform_fields(fields_in) == []
    assert consumed_fields(fields_in, transformer) == []
    request = transformer.transform_request(
        HttpRequest(body={"x": "value"}),
        fields_in,
        fields_out,
        {"x": 0, "y": 0},
    )
    assert request == HttpRequest(
        body='{"x": "value"}',
        headers=Headers(Header("Content-Type", "application/json")),
    )



import logging
from dataclasses import dataclass
from typing import Any, Sequence
from uuid import uuid4

from adaptix import NameStyle, Retort, name_mapping
from requests import Session

from descanso.http.requests import RequestsClient
from descanso.method_descriptor import MethodBinder
from descanso.method_spec import MethodSpec
from descanso.request import (
    FieldDestination,
    HttpRequest,
    RequestTransformer, FieldIn, FieldOut,
)
from descanso.request_transformers import Body, JsonDump, Method, BodyModelDump
from descanso.response import HttpResponse, ResponseTransformer
from descanso.response_transformers import (
    ErrorRaiser,
    JsonLoad,
    KeepResponse,
    BodyModelLoad,
)
from descanso.signature import make_method_spec
from descanso.jsonrpc import JsonRPCBuilder

DEFAULT_BODY_PARAM = "body"



retort = Retort(
    strict_coercion=False,
    recipe=[
        name_mapping(name_style=NameStyle.CAMEL),
    ],
)

jsonrpc = JsonRPCBuilder(
    request_body_dumper=retort,
    response_body_loader=retort,
)


# client
@dataclass
class Transaction:
    block_hash: str
    block_number: str
    from_: str
    chain_id: str
    transaction_index: str


class MyClient(RequestsClient):
    def __init__(self):
        super().__init__(
            base_url="https://eth.merkle.io/",
            session=Session(),
        )

    @jsonrpc("eth_getTransactionByHash")
    def get_transaction_by_hash(self, body: list[str]) -> Transaction:
        """Get transaction"""

    @jsonrpc("net_version")
    def net_version(self) -> str:
        """Retrieve net version"""

    @jsonrpc("eth_blockNumber")
    def eth_block_number(self) -> str:
        """Retrieve block number"""


logging.basicConfig(level=logging.INFO)
c = MyClient()
transaction = c.get_transaction_by_hash(
    ["0x05f71e1b2cb4f03e547739db15d080fd30c989eda04d37ce6264c5686e0722c9"],
)
print("Transaction: ", transaction)
print("Net version: ", c.net_version())
print("Block number:", c.eth_block_number())



from typing import Any

import pytest
from kiss_headers import Header as KissHeader
from kiss_headers import Headers

from descanso.request import (
    FieldDestination,
    FieldIn,
    FieldOut,
    FileData,
    HttpRequest,
)
from descanso.request_transformers import (
    Body,
    Extra,
    File,
    Header,
    Method,
    Query,
    Skip,
    Url,
)
from tests.request_transformers.utills import consumed_fields


def query_int(i: int) -> int:
    return i + 1


@pytest.fixture
def fields_in():
    return [FieldIn("i", int), FieldIn("s", str), FieldIn("a", Any)]


@pytest.fixture
def data_in():
    return {"i": 1, "s": "hello", "a": "any"}


@pytest.mark.parametrize(
    ("transformer", "consumed", "body", "out"),
    [
        (
            Body("unknown"),
            [],
            None,
            [],
        ),
        (
            Body("i"),
            ["i"],
            1,
            [FieldOut(None, FieldDestination.BODY, int)],
        ),
        (
            Body("s"),
            ["s"],
            "hello",
            [FieldOut(None, FieldDestination.BODY, str)],
        ),
    ],
)
def test_body(transformer, consumed, body, out, fields_in, data_in):
    fields_out = transformer.transform_fields(fields_in)
    assert str(transformer)
    assert consumed_fields(fields_in, transformer) == consumed
    assert fields_out == out
    req = transformer.transform_request(
        HttpRequest(),
        fields_in,
        fields_out,
        data_in,
    )
    assert req == HttpRequest(body=body)


@pytest.mark.parametrize(
    ("transformer", "consumed", "extras", "out"),
    [
        (
            Extra("i"),
            ["i"],
            [("i", 1)],
            [FieldOut("i", FieldDestination.EXTRA, int)],
        ),
        (
            Extra("x", "stub"),
            [],
            [("x", "stub")],
            [FieldOut("x", FieldDestination.EXTRA, str)],
        ),
        (
            Extra("x", "{s}{i}"),
            ["i", "s"],
            [("x", "hello1")],
            [FieldOut("x", FieldDestination.EXTRA, str)],
        ),
        (
            Extra("y", query_int),
            ["i"],
            [("y", 2)],
            [FieldOut("y", FieldDestination.EXTRA, int)],
        ),
    ],
)
def test_extra(transformer, consumed, extras, out, fields_in, data_in):
    fields_out = transformer.transform_fields(fields_in)
    assert str(transformer)
    assert consumed_fields(fields_in, transformer) == consumed
    assert fields_out == out
    req = transformer.transform_request(
        HttpRequest(),
        fields_in,
        fields_out,
        data_in,
    )
    assert req == HttpRequest(extras=extras)


@pytest.mark.parametrize(
    ("transformer", "consumed", "headers", "out"),
    [
        (
            Header("i"),
            ["i"],
            Headers(KissHeader("i", "1")),
            [FieldOut("i", FieldDestination.HEADER, int)],
        ),
        (
            Header("x", "stub"),
            [],
            Headers(KissHeader("x", "stub")),
            [FieldOut("x", FieldDestination.HEADER, str)],
        ),
        (
            Header("x", "{s}{i}"),
            ["i", "s"],
            Headers(KissHeader("x", "hello1")),
            [FieldOut("x", FieldDestination.HEADER, str)],
        ),
        (
            Header("y", query_int),
            ["i"],
            Headers(KissHeader("y", "2")),
            [FieldOut("y", FieldDestination.HEADER, int)],
        ),
    ],
)
def test_header(transformer, consumed, headers, out, fields_in, data_in):
    fields_out = transformer.transform_fields(fields_in)
    assert str(transformer)
    assert consumed_fields(fields_in, transformer) == consumed
    assert fields_out == out
    req = transformer.transform_request(
        HttpRequest(),
        fields_in,
        fields_out,
        data_in,
    )
    assert req == HttpRequest(headers=headers)


@pytest.mark.parametrize(
    ("transformer", "consumed", "params", "out"),
    [
        (
            Query("i"),
            ["i"],
            [("i", 1)],
            [FieldOut("i", FieldDestination.QUERY, int)],
        ),
        (
            Query("x", "stub"),
            [],
            [("x", "stub")],
            [FieldOut("x", FieldDestination.QUERY, str)],
        ),
        (
            Query("x", "{s}{i}"),
            ["i", "s"],
            [("x", "hello1")],
            [FieldOut("x", FieldDestination.QUERY, str)],
        ),
        (
            Query("y", query_int),
            ["i"],
            [("y", 2)],
            [FieldOut("y", FieldDestination.QUERY, int)],
        ),
    ],
)
def test_query(transformer, consumed, params, out, fields_in, data_in):
    fields_out = transformer.transform_fields(fields_in)
    assert str(transformer)
    assert consumed_fields(fields_in, transformer) == consumed
    assert fields_out == out
    req = transformer.transform_request(
        HttpRequest(),
        fields_in,
        fields_out,
        data_in,
    )
    assert req == HttpRequest(query_params=params)


@pytest.mark.parametrize(
    ("transformer", "consumed", "out"),
    [
        (
            Skip(),
            [],
            [],
        ),
        (
            Skip("s"),
            ["s"],
            [],
        ),
    ],
)
def test_skip(transformer, consumed, out, fields_in, data_in):
    fields_out = transformer.transform_fields(fields_in)
    assert str(transformer)
    assert consumed_fields(fields_in, transformer) == consumed
    assert fields_out == out
    req = transformer.transform_request(
        HttpRequest(),
        fields_in,
        fields_out,
        data_in,
    )
    assert req == HttpRequest()


@pytest.mark.parametrize(
    ("template", "url", "consumed"),
    [
        ("/", "/", []),
        ("/{i}/{s}", "/1/hello", ["i", "s"]),
        (lambda a: f"/{a}", "/any", ["a"]),
    ],
)
def test_url(template, url, consumed, fields_in, data_in):
    transformer = Url(template)
    fields_out = transformer.transform_fields(fields_in)
    assert str(transformer)
    assert consumed_fields(fields_in, transformer) == consumed
    assert fields_out == [FieldOut(None, FieldDestination.URL, str)]
    req = transformer.transform_request(
        HttpRequest(),
        fields_in,
        fields_out,
        data_in,
    )
    assert req == HttpRequest(url=url)


@pytest.mark.parametrize(
    ("transformer", "consumed", "files", "out"),
    [
        (
            File("s"),
            ["s"],
            [("s", FileData("hello", filename=None, content_type=None))],
            [FieldOut("s", FieldDestination.FILE, str)],
        ),
        (
            File("s", filefield="file"),
            ["s"],
            [("file", FileData("hello", filename=None, content_type=None))],
            [FieldOut("file", FieldDestination.FILE, str)],
        ),
        (
            File("s", filefield="file", filename="a.b"),
            ["s"],
            [("file", FileData("hello", filename="a.b", content_type=None))],
            [FieldOut("file", FieldDestination.FILE, str)],
        ),
        (
            File(
                "s",
                filefield="file",
                filename="a.b",
                content_type="text/plain",
            ),
            ["s"],
            [
                (
                    "file",
                    FileData(
                        "hello",
                        filename="a.b",
                        content_type="text/plain",
                    ),
                ),
            ],
            [FieldOut("file", FieldDestination.FILE, str)],
        ),
    ],
)
def test_file(transformer, consumed, files, out, fields_in, data_in):
    fields_out = transformer.transform_fields(fields_in)
    assert str(transformer)
    assert consumed_fields(fields_in, transformer) == consumed
    assert fields_out == out
    req = transformer.transform_request(
        HttpRequest(),
        fields_in,
        fields_out,
        data_in,
    )
    assert req == HttpRequest(files=files)


def test_pipe(fields_in, data_in):
    query_transformer = Query("i")
    query_transformer2 = Query("s")
    body_transformer = Body("i")
    transformer = query_transformer | query_transformer2 | body_transformer
    fields_out = transformer.transform_fields(fields_in)
    assert str(transformer)
    assert fields_out == [
        FieldOut("i", FieldDestination.QUERY, int),
        FieldOut("s", FieldDestination.QUERY, str),
        FieldOut(None, FieldDestination.BODY, int),
    ]
    assert consumed_fields(fields_in, query_transformer) == ["i"]
    assert consumed_fields(fields_in, query_transformer2) == ["s"]
    assert consumed_fields(fields_in, body_transformer) == ["i"]
    req = transformer.transform_request(
        HttpRequest(),
        fields_in,
        fields_out,
        data_in,
    )
    assert req == HttpRequest(
        body=1,
        query_params=[
            ("i", 1),
            ("s", "hello"),
        ],
    )


def test_method(fields_in, data_in):
    transformer = Method("GET")
    fields_out = transformer.transform_fields(fields_in)
    assert str(transformer)
    assert consumed_fields(fields_in, transformer) == []
    assert fields_out == []
    req = transformer.transform_request(
        HttpRequest(),
        fields_in,
        fields_out,
        data_in,
    )
    assert req == HttpRequest(method="GET")



import asyncio
import concurrent.futures
from threading import Thread

import pytest

from .server import new_site


class MyRunner:
    loop: asyncio.AbstractEventLoop
    future: asyncio.Future

    def __init__(self, port: int) -> None:
        self.port = port
        self.ready = concurrent.futures.Future()

    async def arun(self):
        self.loop = asyncio.get_event_loop()
        self.future = self.loop.create_future()
        site = await new_site(8080)
        await site.start()
        self.ready.set_result(True)
        await self.future
        await site.stop()

    def run(self):
        asyncio.run(self.arun())

    def stop(self):
        self.loop.call_soon_threadsafe(self.future.set_result, None)


@pytest.fixture(scope="module")
def server_addr():
    runner = MyRunner(8080)
    t = Thread(target=runner.run)
    t.start()
    runner.ready.result()
    yield "http://127.0.0.1:8080"
    runner.stop()
    t.join()



__all__ = [
    "AsyncHttpxClient",
    "HttpxClient",
    "HttpxResponseWrapper",
]

import urllib.parse
from collections.abc import AsyncIterator, Iterator, Sequence
from contextlib import asynccontextmanager, contextmanager
from typing import IO, Any

from httpx import AsyncClient as _AsyncClient
from httpx import Client as _Client
from httpx import QueryParams, Response
from kiss_headers import parse_it

from descanso.client import (
    AsyncClient,
    AsyncResponseWrapper,
    SyncClient,
    SyncResponseWrapper,
)
from descanso.request import (
    FileData,
    HttpRequest,
    KeyValueList,
    RequestTransformer,
)
from descanso.response import ResponseTransformer
from descanso.utils import ensure_trailing_slash

_FileName = str | None
_FileContent = IO[bytes] | bytes | str
_FileContentType = str | None
_HttpxFile = (
    _FileContent
    | tuple[_FileName, _FileContent]
    | tuple[_FileName, _FileContent, _FileContentType]
)


class HttpxResponseWrapper(SyncResponseWrapper, AsyncResponseWrapper):
    def __init__(self, response: Response) -> None:
        self.status_code = response.status_code
        self.status_text = response.reason_phrase
        self.url = str(response.url)
        self.headers = parse_it(response.headers)
        self.body = None
        self._raw_response = response

    def load_body(self) -> None:
        self.body = self._raw_response.content

    async def aload_body(self) -> None:
        self.body = self._raw_response.content


def to_httpx_query_params(params: KeyValueList[Any]) -> QueryParams:
    httpx_params = []

    for key, value in params:
        if value is None:
            continue
        httpx_params.append((key, value))

    return QueryParams(httpx_params)


def to_httpx_files(files: KeyValueList[FileData]) -> KeyValueList[_HttpxFile]:
    httpx_files: KeyValueList[_HttpxFile] = []

    for key, file in files:
        if file.contents is None:
            continue

        file_payload = (
            file.filename,
            file.contents,
            file.content_type,
        )

        httpx_files.append((key, file_payload))

    return httpx_files


class HttpxClient(SyncClient):
    def __init__(
        self,
        base_url: str,
        session: _Client,
        transformers: Sequence[RequestTransformer | ResponseTransformer] = (),
    ) -> None:
        super().__init__(
            transformers=transformers,
        )

        self._base_url = ensure_trailing_slash(base_url)
        self._session = session

    @contextmanager
    def send_request(
        self,
        request: HttpRequest,
    ) -> Iterator[SyncResponseWrapper]:
        response = self._session.request(
            method=request.method,
            url=urllib.parse.urljoin(self._base_url, request.url),
            headers=request.headers.items(),
            data=request.body,
            params=to_httpx_query_params(request.query_params),
            files=to_httpx_files(request.files),
        )

        yield HttpxResponseWrapper(response)


class AsyncHttpxClient(AsyncClient):
    def __init__(
        self,
        base_url: str,
        session: _AsyncClient,
        transformers: Sequence[RequestTransformer | ResponseTransformer] = (),
    ) -> None:
        super().__init__(
            transformers=transformers,
        )

        self._base_url = ensure_trailing_slash(base_url)
        self._session = session

    @asynccontextmanager
    async def asend_request(
        self,
        request: HttpRequest,
    ) -> AsyncIterator[AsyncResponseWrapper]:
        response = await self._session.request(
            method=request.method,
            url=urllib.parse.urljoin(self._base_url, request.url),
            headers=request.headers.items(),
            data=request.body,
            params=to_httpx_query_params(request.query_params),
            files=to_httpx_files(request.files),
        )

        yield HttpxResponseWrapper(response)
